@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003 GraphicsMagick Group
+% Copyright (C) 2003-2020 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -36,11 +36,15 @@
   Include declarations.
 */
 #include "magick/studio.h"
+#include "magick/analyze.h"
 #include "magick/blob.h"
-#include "magick/cache.h"
 #include "magick/color.h"
+#include "magick/color_lookup.h"
+#include "magick/colormap.h"
+#include "magick/log.h"
 #include "magick/magick.h"
 #include "magick/monitor.h"
+#include "magick/pixel_cache.h"
 #include "magick/quantize.h"
 #include "magick/resize.h"
 #include "magick/utility.h"
@@ -123,12 +127,66 @@ static unsigned int IsXPM(const unsigned char *magick,const size_t length)
 %
 */
 
+static char **StringToListMod(char *text)
+{
+  char
+    **textlist;
+
+  register char
+    *p;
+
+  register size_t
+    i;
+
+  size_t
+    lines;
+
+  if (text == (char *) NULL)
+    return((char **) NULL);
+
+  /*
+    Convert string to an ASCII list, modifying the input string.
+  */
+  lines=1;
+  for (p=text; *p != '\0'; p++)
+    if (*p == '\n')
+      lines++;
+  textlist=MagickAllocateMemory(char **,(lines+1)*sizeof(char *));
+  if (textlist == (char **) NULL)
+    MagickFatalError3(ResourceLimitFatalError,MemoryAllocationFailed,
+                      UnableToConvertText);
+  i=0;
+  textlist[i]=text;
+  for (p=text; *p != '\0'; p++)
+    {
+      if (*p == '\r')
+        {
+          *p='\0';
+          continue;
+        }
+      if (*p == '\n')
+        {
+          *p='\0';
+          if (*(p+1) != '\0')
+            {
+              i++;
+              textlist[i]=p+1;
+            }
+        }
+    }
+  i++;
+  textlist[i]=(char *) NULL;
+  return(textlist);
+}
+
+static char *ParseColor(char *data) MAGICK_FUNC_PURE;
+
 static char *ParseColor(char *data)
 {
 #define NumberTargets  6
 
   static const char
-    *targets[NumberTargets] = { "c ", "g ", "g4 ", "m ", "b ", "s " };
+    targets[NumberTargets][3] = { "c ", "g ", "g4 ", "m ", "b ", "s " };
 
   register char
     *p,
@@ -162,15 +220,25 @@ static char *ParseColor(char *data)
   }
   return((char *) NULL);
 }
+#define ThrowXPMReaderException(code_,reason_,image_) \
+do { \
+  if (keys) \
+    for (i=0; i < (long) image->colors; i++) \
+      MagickFreeMemory(keys[i]); \
+  MagickFreeMemory(keys); \
+  MagickFreeMemory(textlist); \
+  MagickFreeMemory(xpm_buffer); \
+  ThrowReaderException(code_,reason_,image_); \
+} while (0);
 
 static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
 {
   char
     key[MaxTextExtent],
-    **keys,
+    **keys = (char **) NULL,
     target[MaxTextExtent],
-    **textlist,
-    *xpm_buffer;
+    **textlist = (char **) NULL,
+    *xpm_buffer = (char *) NULL;
 
   Image
     *image;
@@ -178,9 +246,12 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   int
     count;
 
-  long
+  unsigned long
     j,
-    none,
+    none;
+
+  long
+    k,
     y;
 
   register char
@@ -199,12 +270,17 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   register long
     i;
 
-  unsigned int
+  MagickPassFail
     status;
 
+  MagickBool
+    colormap_initialized;
+
+  size_t
+    length;
+
   unsigned long
-    length,
-    width;
+    width; /* characters per pixel */
 
   /*
     Open image file.
@@ -215,101 +291,147 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   assert(exception->signature == MagickSignature);
   image=AllocateImage(image_info);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
-  if (status == False)
-    ThrowReaderException(FileOpenError,UnableToOpenFile,image);
+  if (status == MagickFail)
+    ThrowXPMReaderException(FileOpenError,UnableToOpenFile,image);
   /*
     Read XPM file.
   */
   length=MaxTextExtent;
   xpm_buffer=MagickAllocateMemory(char *,length);
-  p=xpm_buffer;
   if (xpm_buffer != (char *) NULL)
-    while (ReadBlobString(image,p) != (char *) NULL)
     {
-      if (*p == '#')
-        if ((p == xpm_buffer) || (*(p-1) == '\n'))
-          continue;
-      if ((*p == '}') && (*(p+1) == ';'))
-        break;
-      p+=strlen(p);
-      if ((p-xpm_buffer+MaxTextExtent+1) < (long) length)
-        continue;
-      length<<=1;
-      MagickReallocMemory(xpm_buffer,length);
-      if (xpm_buffer == (char *) NULL)
-        break;
-      p=xpm_buffer+strlen(xpm_buffer);
+      xpm_buffer[0]='\0';
+      p=xpm_buffer;
+      while (ReadBlobString(image,p) != (char *) NULL)
+        {
+          if (*p == '#')
+            if ((p == xpm_buffer) || (*(p-1) == '\n'))
+              continue;
+          if ((*p == '}') && (*(p+1) == ';'))
+            break;
+          p+=strlen(p);
+          if (((size_t) (p-xpm_buffer)+MaxTextExtent+1) < length)
+            continue;
+          length<<=1;
+          MagickReallocMemory(char *,xpm_buffer,length);
+          if (xpm_buffer == (char *) NULL)
+            break;
+          p=xpm_buffer+strlen(xpm_buffer);
+        }
     }
   if (xpm_buffer == (char *) NULL)
-    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+    ThrowXPMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
   /*
-    Remove comments.
+    Parse image properties from file header while skipping over comment markers.
   */
   count=0;
-  for (p=xpm_buffer; *p != '\0'; p++)
+  for (p=xpm_buffer; ((p - xpm_buffer) < 512) && (*p != '\0'); p++)
   {
-    if (*p != '"')
+    if ((*p != '"') || !isdigit((int) *(p+1)))
       continue;
-    count=sscanf(p+1,"%lu %lu %lu %lu",&image->columns,&image->rows,
+    count=sscanf(p+1,"%lu %lu %u %lu",&image->columns,&image->rows,
       &image->colors,&width);
     if (count == 4)
-      break;
+      {
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Columns: %lu, Rows: %lu, Colors: %u, Char Per Pixel: %lu",
+                              image->columns, image->rows, image->colors, width);
+        break;
+      }
   }
-  if ((count != 4) || (width > 2) || (image->columns == 0) ||
-      (image->rows == 0) || (image->colors == 0))
-    ThrowReaderException(CorruptImageError,ImproperImageHeader,image)
+  if ((count != 4) || (width == 0) || (width > 2) ||
+      (image->columns == 0) || (image->rows == 0) ||
+      (image->colors == 0) || (image->colors > MaxColormapSize))
+    ThrowXPMReaderException(CorruptImageError,ImproperImageHeader,image);
   image->depth=16;
   /*
     Remove unquoted characters.
   */
-  i=0;
-  for ( ; *p != '\0'; p++)
   {
-    if (*p != '"')
-      continue;
-    for (q=p+1; *q != '\0'; q++)
-      if (*q == '"')
-        break;
-    (void) strncpy(xpm_buffer+i,p+1,q-p-1);
-    i+=q-p-1;
-    xpm_buffer[i++]='\n';
-    p=q+1;
+    MagickBool inquote = MagickFalse;
+    q=xpm_buffer;
+    while (*p != '\0')
+      {
+        if (*p++ == '"')
+          {
+            if (inquote)
+              *q++='\n';
+            inquote = !inquote;
+          }
+        if (inquote)
+          *q++=*p;
+      }
+    *q='\0';
+    if (inquote)
+      {
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Imbalanced quotes error");
+        ThrowXPMReaderException(CorruptImageError,CorruptImage,image);
+      }
   }
-  xpm_buffer[i]='\0';
-  textlist=StringToList(xpm_buffer);
-  MagickFreeMemory(xpm_buffer);
+  /*
+    Scan for non-white space binary control codes and reject file if
+    they are present.
+  */
+  for (p=xpm_buffer; *p != '\0'; p++)
+    if (((unsigned char) *p < 32) && !isspace((int)(unsigned char) (*p)))
+      break;
+  if (*p != '\0')
+    {
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Binary control codes error");
+        ThrowXPMReaderException(CorruptImageError,CorruptImage,image);
+    }
+  textlist=StringToListMod(xpm_buffer);
   if (textlist == (char **) NULL)
-    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+    ThrowXPMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+#if 0
+  if (image->logging)
+    {
+      for (i=0; textlist[i] != (char *) NULL; i++)
+        { };
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                            "TextList has %lu entries", i);
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                            "TextList");
+      for (i=0; textlist[i] != (char *) NULL; i++)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "    %lu: \"%s\"", i, textlist[i]);
+    }
+#endif
+
   /*
     Initialize image structure.
   */
-  keys=MagickAllocateMemory(char **,image->colors*sizeof(char *));
-  if (!AllocateImageColormap(image,image->colors) || (keys == (char **) NULL))
-    {
-      for (i=0; textlist[i] != (char *) NULL; i++)
-        MagickFreeMemory(textlist[i]);
-      MagickFreeMemory(textlist);
-      ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image)
-    }
+  keys=MagickAllocateArray(char **,image->colors,sizeof(char *));
+  if (keys == (char **) NULL)
+    ThrowXPMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+  for (i=0; i < (long) image->colors; i++)
+    keys[i]=(char *) NULL;
+  if (!AllocateImageColormap(image,image->colors))
+    ThrowXPMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+
   /*
     Read image colormap.
   */
   i=1;
-  none=(-1);
-  for (j=0; j < (long) image->colors; j++)
+  none=(~0U);
+  (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                        "Parsing colormap...");
+  colormap_initialized=MagickFalse;
+  for (j=0; j < image->colors; j++)
   {
     p=textlist[i++];
-    if (p == (char *) NULL)
+    if ((p == (char *) NULL) || (p[0] == '\0'))
       break;
-    keys[j]=MagickAllocateMemory(char *,width+1);
+    if (image->logging)
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "    %lu: %s", i-1, textlist[i-1]);
+    if (strlen(p) < width)
+      break;
+    keys[j]=MagickAllocateMemory(char *,(size_t) width+1);
     if (keys[j] == (char *) NULL)
-      {
-        for (i=0; textlist[i] != (char *) NULL; i++)
-          MagickFreeMemory(textlist[i]);
-        MagickFreeMemory(textlist);
-        MagickFreeMemory(keys);
-        ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image)
-      }
+      ThrowXPMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
     keys[j][width]='\0';
     (void) strncpy(keys[j],p,width);
     /*
@@ -321,7 +443,7 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
       {
         while (!isspace((int) (*q)) && (*q != '\0'))
           q++;
-        (void) strncpy(target,q,MaxTextExtent-1);
+        (void) strlcpy(target,q,MaxTextExtent);
         q=ParseColor(target);
         if (q != (char *) NULL)
           *q='\0';
@@ -335,15 +457,19 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
         (void) strcpy(target,"black");
       }
     if (!QueryColorDatabase(target,&image->colormap[j],exception))
-      break;
+      {
+        /* Promote warning to error */
+        exception->severity = CorruptImageError;
+        break;
+      }
+    /* We are going to be done now */
+    if (j+1 == image->colors)
+      colormap_initialized=MagickTrue;
   }
-  if (j < (long) image->colors)
-    {
-      for (i=0; textlist[i] != (char *) NULL; i++)
-        MagickFreeMemory(textlist[i]);
-      MagickFreeMemory(textlist);
-      ThrowReaderException(CorruptImageError,CorruptImage,image)
-    }
+  if (!colormap_initialized)
+    ThrowXPMReaderException(CorruptImageError,CorruptImage,image);
+  image->depth=GetImageDepth(image,&image->exception);
+  image->depth=NormalizeDepthToOctet(image->depth);
   j=0;
   key[width]='\0';
   if (!image_info->ping)
@@ -351,35 +477,55 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
       /*
         Read image pixels.
       */
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                        "Parsing pixels...");
       for (y=0; y < (long) image->rows; y++)
       {
         p=textlist[i++];
-        if (p == (char *) NULL)
+        if ((p == (char *) NULL) || (p[0] == '\0'))
           break;
-        r=SetImagePixels(image,0,y,image->columns,1);
+        if (image->logging)
+          (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                "    %lu: %s", i-1, textlist[i-1]);
+        r=SetImagePixelsEx(image,0,y,image->columns,1,exception);
         if (r == (PixelPacket *) NULL)
           break;
-        indexes=GetIndexes(image);
+        indexes=AccessMutableIndexes(image);
         for (x=0; x < (long) image->columns; x++)
         {
-          (void) strncpy(key,p,width);
+          /* (void) strncpy(key,p,width); */
+          for (k=0; k < (long) width; k++)
+            {
+              key[k]=p[k];
+              if (p[k] == '\0')
+                {
+                  status=MagickFail;
+                  break;
+                }
+            }
+          if (MagickFail == status)
+            break;
+          key[k]='\0';
           if (strcmp(key,keys[j]) != 0)
-            for (j=0; j < (long) Max(image->colors-1,1); j++)
+            for (j=0; j < Max(image->colors-1,1); j++)
               if (strcmp(key,keys[j]) == 0)
                 break;
+          VerifyColormapIndex(image,j);
           if (image->storage_class == PseudoClass)
             indexes[x]=(IndexPacket) j;
           *r=image->colormap[j];
           r->opacity=(Quantum)
-            (j == (long) none ? TransparentOpacity : OpaqueOpacity);
+            (j == none ? TransparentOpacity : OpaqueOpacity);
           r++;
           p+=width;
         }
-        if (!SyncImagePixels(image))
+        if (MagickFail == status)
+          break;
+        if (!SyncImagePixelsEx(image,exception))
           break;
       }
       if (y < (long) image->rows)
-        ThrowReaderException(CorruptImageError,InsufficientImageDataInFile,image);
+        ThrowXPMReaderException(CorruptImageError,InsufficientImageDataInFile,image);
     }
   /*
     Free resources.
@@ -387,10 +533,10 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   for (i=0; i < (long) image->colors; i++)
     MagickFreeMemory(keys[i]);
   MagickFreeMemory(keys);
-  for (i=0; textlist[i] != (char *) NULL; i++)
-    MagickFreeMemory(textlist[i]);
   MagickFreeMemory(textlist);
+  MagickFreeMemory(xpm_buffer);
   CloseBlob(image);
+  StopTimer(&image->timer);
   return(image);
 }
 
@@ -426,8 +572,8 @@ ModuleExport void RegisterXPMImage(void)
   entry->decoder=(DecoderHandler) ReadXPMImage;
   entry->encoder=(EncoderHandler) WritePICONImage;
   entry->adjoin=False;
-  entry->description=AcquireString("Personal Icon");
-  entry->module=AcquireString("XPM");
+  entry->description="Personal Icon";
+  entry->module="XPM";
   (void) RegisterMagickInfo(entry);
 
   entry=SetMagickInfo("PM");
@@ -435,8 +581,8 @@ ModuleExport void RegisterXPMImage(void)
   entry->encoder=(EncoderHandler) WriteXPMImage;
   entry->adjoin=False;
   entry->stealth=True;
-  entry->description=AcquireString("X Windows system pixmap (color)");
-  entry->module=AcquireString("XPM");
+  entry->description="X Windows system pixmap (color)";
+  entry->module="XPM";
   (void) RegisterMagickInfo(entry);
 
   entry=SetMagickInfo("XPM");
@@ -444,8 +590,8 @@ ModuleExport void RegisterXPMImage(void)
   entry->encoder=(EncoderHandler) WriteXPMImage;
   entry->magick=(MagickHandler) IsXPM;
   entry->adjoin=False;
-  entry->description=AcquireString("X Windows system pixmap (color)");
-  entry->module=AcquireString("XPM");
+  entry->description="X Windows system pixmap (color)";
+  entry->module="XPM";
   (void) RegisterMagickInfo(entry);
 }
 
@@ -511,7 +657,7 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
 #define GraymapExtent  95
 #define PiconGeometry  "48x48>"
 
-  static unsigned char
+  static const unsigned char
     Colormap[]=
     {
       0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x06, 0x00, 0x05, 0x00, 0xf4, 0x05,
@@ -569,9 +715,6 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
   register const PixelPacket
     *p;
 
-  register IndexPacket
-    *indexes;
-
   register long
     i,
     x;
@@ -587,24 +730,39 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
     characters_per_pixel,
     colors;
 
-  /*
-    Open output image file.
-  */
+    ImageCharacteristics
+      characteristics;
+
   assert(image_info != (const ImageInfo *) NULL);
   assert(image_info->signature == MagickSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
+  /*
+    Open output image file.
+  */
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == False)
     ThrowWriterException(FileOpenError,UnableToOpenFile,image);
-  TransformColorspace(image,RGBColorspace);
+  /*
+    Ensure that image is in an RGB space.
+  */
+  (void) TransformColorspace(image,RGBColorspace);
+  /*
+    Analyze image to be written.
+  */
+  if (!GetImageCharacteristics(image,&characteristics,
+                               (OptimizeType == image_info->type),
+                               &image->exception))
+    {
+      CloseBlob(image);
+      return MagickFail;
+    }
   SetGeometry(image,&geometry);
   (void) GetMagickGeometry(PiconGeometry,&geometry.x,&geometry.y,
     &geometry.width,&geometry.height);
-  picon=ResizeImage(image,geometry.width,geometry.height,TriangleFilter,1.0,
-    &image->exception);
+  picon=ThumbnailImage(image,geometry.width,geometry.height,&image->exception);
   if ((image_info->type != TrueColorType) &&
-      IsGrayImage(image,&image->exception))
+      (characteristics.grayscale))
     map=BlobToImage(image_info,Graymap,GraymapExtent,&image->exception);
   else
     map=BlobToImage(image_info,Colormap,ColormapExtent,&image->exception);
@@ -646,19 +804,26 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
               break;
           }
         }
-      SetImageType(picon,PaletteType);
+      (void) SetImageType(picon,PaletteType);
     }
   colors=picon->colors;
   if (transparent)
     {
       colors++;
-      MagickReallocMemory(picon->colormap,colors*sizeof(PixelPacket));
+      MagickReallocMemory(PixelPacket *,picon->colormap,colors*sizeof(PixelPacket));
+      if (picon->colormap == (PixelPacket *) NULL)
+        ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,picon);
+      picon->colormap[colors-1].red=picon->colormap[colors-1].green=picon->colormap[colors-1].blue=0;
+      picon->colormap[colors-1].opacity=TransparentOpacity;
       for (y=0; y < (long) picon->rows; y++)
       {
+        register IndexPacket
+          *indexes;
+
         q=GetImagePixels(picon,0,y,picon->columns,1);
         if (q == (PixelPacket *) NULL)
           break;
-        indexes=GetIndexes(picon);
+        indexes=AccessMutableIndexes(picon);
         for (x=0; x < (long) picon->columns; x++)
         {
           if (q->opacity == TransparentOpacity)
@@ -719,10 +884,13 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
   (void) WriteBlobString(image,"/* pixels */\n");
   for (y=0; y < (long) picon->rows; y++)
   {
+    register const IndexPacket
+      *indexes;
+
     p=AcquireImagePixels(picon,0,y,picon->columns,1,&picon->exception);
     if (p == (const PixelPacket *) NULL)
       break;
-    indexes=GetIndexes(picon);
+    indexes=AccessImmutableIndexes(picon);
     (void) WriteBlobString(image,"\"");
     for (x=0; x < (long) picon->columns; x++)
     {
@@ -734,14 +902,16 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
         symbol[j]=Cixel[k];
       }
       symbol[j]='\0';
-      (void) strncpy(buffer,symbol,MaxTextExtent-1);
+      (void) strlcpy(buffer,symbol,MaxTextExtent);
       (void) WriteBlobString(image,buffer);
     }
     FormatString(buffer,"\"%.1024s\n",
       (y == (long) (picon->rows-1) ? "" : ","));
     (void) WriteBlobString(image,buffer);
     if (QuantumTick(y,picon->rows))
-      if (!MagickMonitor(SaveImageText,y,picon->rows,&image->exception))
+      if (!MagickMonitorFormatted(y,picon->rows,&image->exception,
+                                  SaveImageText,image->filename,
+                                  image->columns,image->rows))
         break;
   }
   DestroyImage(picon);
@@ -803,9 +973,6 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
   register const PixelPacket
     *p;
 
-  register IndexPacket
-    *indexes;
-
   register long
     i,
     x;
@@ -831,7 +998,12 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == False)
     ThrowWriterException(FileOpenError,UnableToOpenFile,image);
-  TransformColorspace(image,RGBColorspace);
+  (void) TransformColorspace(image,RGBColorspace);
+  /*
+    Limit color resolution to what XPM can traditionally handle
+  */
+  if (image->depth > 16)
+    image->depth=16;
   transparent=False;
   if (image->storage_class == PseudoClass)
     {
@@ -866,19 +1038,26 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
               break;
           }
         }
-      SetImageType(image,PaletteType);
+      (void) SetImageType(image,PaletteType);
     }
   colors=image->colors;
   if (transparent)
     {
       colors++;
-      MagickReallocMemory(image->colormap,colors*sizeof(PixelPacket));
+      MagickReallocMemory(PixelPacket *,image->colormap,colors*sizeof(PixelPacket));
+      if (image->colormap == (PixelPacket *) NULL)
+        ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,image);
+      image->colormap[colors-1].red=image->colormap[colors-1].green=image->colormap[colors-1].blue=0;
+      image->colormap[colors-1].opacity=TransparentOpacity;
       for (y=0; y < (long) image->rows; y++)
       {
+        register IndexPacket
+          *indexes;
+
         q=GetImagePixels(image,0,y,image->columns,1);
         if (q == (PixelPacket *) NULL)
           break;
-        indexes=GetIndexes(image);
+        indexes=AccessMutableIndexes(image);
         for (x=0; x < (long) image->columns; x++)
         {
           if (q->opacity == TransparentOpacity)
@@ -939,10 +1118,13 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
   (void) WriteBlobString(image,"/* pixels */\n");
   for (y=0; y < (long) image->rows; y++)
   {
+    register const IndexPacket
+      *indexes;
+
     p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
     if (p == (const PixelPacket *) NULL)
       break;
-    indexes=GetIndexes(image);
+    indexes=AccessImmutableIndexes(image);
     (void) WriteBlobString(image,"\"");
     for (x=0; x < (long) image->columns; x++)
     {
@@ -954,14 +1136,16 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
         symbol[j]=Cixel[k];
       }
       symbol[j]='\0';
-      (void) strncpy(buffer,symbol,MaxTextExtent-1);
+      (void) strlcpy(buffer,symbol,MaxTextExtent);
       (void) WriteBlobString(image,buffer);
     }
     FormatString(buffer,"\"%.1024s\n",
       (y == (long) (image->rows-1) ? "" : ","));
     (void) WriteBlobString(image,buffer);
     if (QuantumTick(y,image->rows))
-      if (!MagickMonitor(SaveImageText,y,image->rows,&image->exception))
+      if (!MagickMonitorFormatted(y,image->rows,&image->exception,
+                                  SaveImageText,image->filename,
+                                  image->columns,image->rows))
         break;
   }
   (void) WriteBlobString(image,"};\n");
